import pytest
import os
from agogosml.common.eventhub_streaming_client import EventHubStreamingClient
from agogosml.common.kafka_streaming_client import KafkaStreamingClient
from agogosml.reader.input_reader import InputReader
from agogosml.reader.input_reader_factory import InputReaderFactory
from .client_mocks import StreamingClientMock, MessageSenderMock


@pytest.fixture
def mock_streaming_client():
    return StreamingClientMock()


@pytest.fixture
def mock_message_sender():
    return MessageSenderMock()


def test_when_instance_created(mock_streaming_client, mock_message_sender):
    ir = InputReader(mock_streaming_client, mock_message_sender)
    assert ir is not None


def test_when_start_receiving_then_messaging_client_starts(mock_streaming_client, mock_message_sender):
    # arrange
    ir = InputReader(mock_streaming_client, mock_message_sender)

    # act
    ir.start_receiving_messages()
    assert mock_streaming_client.get_receiving()


def test_when_msg_received_callback_called(mock_streaming_client,
                                           mock_message_sender):
    # arrange
    ir = InputReader(mock_streaming_client, mock_message_sender)

    # act
    ir.start_receiving_messages()
    mock_streaming_client.fake_incoming_message_from_streaming('a')
    assert mock_message_sender.get_last_msg() == 'a'


def test_when_failed_to_deliver_then_failure_callback(mock_streaming_client, mock_message_sender):
    # arrange
    ir = InputReader(mock_streaming_client, mock_message_sender)

    # act
    ir.start_receiving_messages()
    mock_message_sender.set_fail_send(True)
    result = mock_streaming_client.fake_incoming_message_from_streaming('a')

    # assert
    assert result is False


# TODO: Try splitting out the unit tests and integration tests into two different directories within tests


@pytest.mark.integration
def test_eventhub_created_from_factory():
    config = {
        'client': {
            'type': 'eventhub',
            'config': {
                "AZURE_STORAGE_ACCOUNT": os.getenv("AZURE_STORAGE_ACCOUNT"),
                "AZURE_STORAGE_ACCESS_KEY": os.getenv("AZURE_STORAGE_ACCESS_KEY"),
                "LEASE_CONTAINER_NAME": os.getenv("LEASE_CONTAINER_NAME"),
                "EVENT_HUB_NAMESPACE": os.getenv("EVENT_HUB_NAMESPACE"),
                "EVENT_HUB_NAME": os.getenv("EVENT_HUB_NAME"),
                "EVENT_HUB_SAS_POLICY": os.getenv("EVENT_HUB_SAS_POLICY"),
                "EVENT_HUB_SAS_KEY": os.getenv("EVENT_HUB_SAS_KEY"),
                "EVENT_HUB_CONSUMER_GROUP": os.getenv("EVENT_HUB_CONSUMER_GROUP"),
                "TIMEOUT": 60
            }
        },
        "APP_HOST": os.getenv("APP_HOST"),
        "APP_PORT": os.getenv("APP_PORT")
    }

    ir = InputReaderFactory.create(config)
    assert ir is not None
    assert isinstance(ir, InputReader)
    assert isinstance(ir.messaging_client, EventHubStreamingClient)


@pytest.mark.integration
def test_kafka_created_from_factory():
    config = {
        'client': {
            'type': 'kafka',
            'config': {
                "KAFKA_TOPIC": os.getenv("KAFKA_TOPIC"),
                "KAFKA_ADDRESS": os.getenv("KAFKA_ADDRESS"),
                "KAFKA_CONSUMER_GROUP": os.getenv("KAFKA_CONSUMER_GROUP"),
                "TIMEOUT": 20
            }
        },
        "APP_HOST": os.getenv("APP_HOST"),
        "APP_PORT": os.getenv("APP_PORT")
    }
    ir = InputReaderFactory.create(config)
    assert ir is not None
    assert isinstance(ir, InputReader)
    assert isinstance(ir.messaging_client, KafkaStreamingClient)



from agogosml.common.flask_http_listener_client import FlaskHttpListenerClient
from agogosml.common.http_message_sender import HttpMessageSender
from agogosml.utils.logger import Logger

logger = Logger()


class TestApp:

    def __init__(self, app_port, app_host, output_port, output_host):
        self.listener = FlaskHttpListenerClient(app_port, app_host)
        self.sender = HttpMessageSender(output_host, output_port)

    def start(self):
        self.listener.start(self.on_message_received)

    def on_message_received(self, message):
        logger.info('Test App message received: ' + message)
        sender_result = self.sender.send(message)
        logger.info('Sender result: ' + str(sender_result))
        return sender_result



"""Abstract Message Sender class"""

from abc import ABC, abstractmethod


class MessageSender(ABC):  # pylint: disable=too-few-public-methods
    """Message Sender class."""

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def send(self, *args, **kwargs):
        """
        Sends message.
        """
        pass



"""Abstract streaming client class"""

from abc import ABC, abstractmethod


class AbstractStreamingClient(ABC):
    """Abstract Streaming Client"""
    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def send(self, *args, **kwargs):
        """Send method."""
        pass

    @abstractmethod
    def stop(self, *args, **kwargs):
        """Stop method."""
        pass

    @abstractmethod
    def start_receiving(self, *args, **kwargs):
        """Start receiving messages from streaming client."""
        pass



# -*- coding: utf-8 -*-
""" InputReader """

from agogosml.common.abstract_streaming_client import AbstractStreamingClient
from agogosml.common.message_sender import MessageSender
from agogosml.utils.logger import Logger

logger = Logger()


class InputReader:  # pylint: disable=too-few-public-methods
    """Accepts incoming messages and routes them to a configured output"""

    def __init__(self, streaming_client: AbstractStreamingClient, message_sender: MessageSender):
        """
        :param streaming_client: A client that can stream data out
        :param message_sender: A message sender that can send messages onwards
        """
        self.message_sender = message_sender
        self.messaging_client = streaming_client

    def start_receiving_messages(self):
        """Start receiving messages from streaming endpoint"""
        self.messaging_client.start_receiving(self.on_message_received)

    def stop_incoming_messages(self):
        """Stop incoming messages from streaming endpoint"""
        self.messaging_client.stop()

    def on_message_received(self, message):
        """Send messages onwards

        :param message: a message to process
        """
        result = self.message_sender.send(message)
        if result:
            return True
            pass
        else:
            self.handle_send_failure(message)
            return False

    def handle_send_failure(self, message):
        logger.error('Error while sending message to App')
        # Notify...



from agogosml.common.abstract_streaming_client import AbstractStreamingClient
from agogosml.common.listener_client import ListenerClient
from agogosml.common.message_sender import MessageSender


class StreamingClientMock(AbstractStreamingClient):
    """
    A class to mock functionality at the streaming client level.
    """
    def __init__(self):
        self.sent = False
        self.receiving = False
        self.last_message = None
        self.should_fail_to_send = False
        pass

    def send(self, msg):
        print('Streaming Client Mock send message: '+msg)
        if self.should_fail_to_send:
            self.sent = False
            return False
        else:
            self.sent = True
            self.last_message = msg
            return True

    def stop(self, *args, **kwargs):
        pass

    def start_receiving(self, callback):
        self.receiving = True
        self.callback = callback
        pass

    def get_sent(self):
        return self.sent

    def get_last_msg(self):
        return self.last_message

    def get_receiving(self):
        return self.receiving

    def fake_incoming_message_from_streaming(self, msg):
        return self.callback(msg)

    def set_fail_send(self, should_fail):
        self.should_fail_to_send = should_fail


class HttpClientMock(ListenerClient):
    """
    A class to mock functionality at the http client level
    """
    def __init__(self, port):
        self.callback = None
        self.startCalled = False
        self.stopCalled = False

    def start(self, on_message_received):
        self.callback = on_message_received
        self.startCalled = True
        pass

    def stop(self):
        self.stopCalled = True
        pass

    def mock_new_incoming_message(self):
        return self.callback("{'some':'json'}")

    def get_started(self):
        return self.startCalled

    def get_stopped(self):
        return self.stopCalled


class MessageSenderMock(MessageSender):
    def __init__(self):
        self.msg = None
        self.should_fail_to_send = None
        pass

    def send(self, msg):
        if self.should_fail_to_send:
            return False
        else:
            self.msg = msg
            return True

    def get_last_msg(self):
        return self.msg

    def set_fail_send(self, should_fail):
        self.should_fail_to_send = should_fail



""" Flask client to receive messages from customer app"""
import threading
from flask import Flask, request
from .listener_client import ListenerClient

DEFAULT_HOST = '127.0.0.1'


class FlaskHttpListenerClient(ListenerClient):
    def __init__(self, port, host=DEFAULT_HOST):
        self.port = port
        self.host = host

    def thread_flask(self):
        app = Flask(__name__)

        @app.route("/", methods=["POST"])
        def on_input():
            msg = str(request.data, 'utf-8', 'ignore')
            if self.on_message_received(msg):
                return 'msg:' + msg
            else:
                print('Error: The callback failed to process the message, returning 500')
                return 'Error: The callback failed to process the message', 500

        app.run(
            host=self.host, port=self.port, debug=False, use_reloader=False, threaded=True)

    def start(self, on_message_received):
        self.on_message_received = on_message_received
        self.t_flask = threading.Thread(name='agogosml', target=self.thread_flask)
        self.t_flask.setDaemon(True)
        self.t_flask.start()

    def stop(self):
        self.shutdown_server()
        # self.t_flask.join()

    def shutdown_server(self):
        try:
            func = request.environ.get('werkzeug.server.shutdown')
            if func is None:
                raise RuntimeError('Not running with the Werkzeug Server')
            func()
        except Exception as e:
            print('error while shutting down flask server: ' + str(e))



import datetime
import threading

import pytest
import os
import time
from dotenv import load_dotenv

from agogosml.common.kafka_streaming_client import KafkaStreamingClient
from agogosml.reader.input_reader_factory import InputReaderFactory
from agogosml.writer.output_writer_factory import OutputWriterFactory
from tests.client_mocks import StreamingClientMock, HttpClientMock
from tests.integration_tests.test_app import TestApp

load_dotenv()


@pytest.fixture
def mock_streaming_client():
    return StreamingClientMock()


@pytest.fixture
def mock_listener_client():
    return HttpClientMock(0)


@pytest.mark.integration
def test_when_messages_received_in_input_then_all_messages_are_sent_via_output():
    """
    This function tests the integration of input and output.
    It assumes connectivity to streaming client is correct, therefore the steaming client is mocked.
    We look to find that once a message is received, it is send out by the output writer.
    This is the maximal integration test we can have in an isolated environment.
    :return:
    """

    # setup input
    input_client_mock = StreamingClientMock()
    ir_config = {
        'APP_PORT': os.getenv("APP_PORT"),
        'APP_HOST': os.getenv("APP_HOST"),
    }
    print('Creating reader')
    ir = InputReaderFactory.create(ir_config, input_client_mock)
    print('start_receiving_messages')
    ir.start_receiving_messages()

    # setup app
    app = TestApp(os.getenv("APP_PORT"),
                  os.getenv("APP_HOST"),
                  os.getenv("OUTPUT_WRITER_PORT"),
                  os.getenv("OUTPUT_WRITER_HOST"))
    app.start()

    # setup output
    ow_config = {
        'OUTPUT_WRITER_PORT': os.getenv("OUTPUT_WRITER_PORT"),
        'OUTPUT_WRITER_HOST': os.getenv("OUTPUT_WRITER_HOST"),
    }

    print('Creating writer')
    output_client_mock = StreamingClientMock()
    ow = OutputWriterFactory.create(ow_config, output_client_mock, None)
    print('start_incoming_messages')
    ow.start_incoming_messages()

    print('sending test message to reader')
    test_msg = str(time.clock())
    # send a message from INPUT reader, and expect it to flow in the pipeline,
    # and eventually be picked up by the output writer
    input_client_mock.fake_incoming_message_from_streaming(test_msg)
    last_msg = output_client_mock.get_last_msg()

    assert last_msg == test_msg
    ir.stop_incoming_messages()
    ow.stop_incoming_messages()


@pytest.mark.integration
def test_when_error_in_output_then_pipeline_fails():
    """
    This function tests the integration of input and output.
    It assumes connectivity to streaming client is correct, therefore the steaming client is mocked.
    We look to find that once a message is received, it is send out by the output writer.
    This is the maximal integration test we can have in an isolated environment.
    :return:
    """

    # setup input
    input_client_mock = StreamingClientMock()
    ir_config = {
        'APP_PORT': os.getenv("APP_PORT"),
        'APP_HOST': os.getenv("APP_HOST"),
    }
    print('Creating reader')
    ir = InputReaderFactory.create(ir_config, input_client_mock)
    print('start_receiving_messages')
    ir.start_receiving_messages()

    # setup app
    app = TestApp(os.getenv("APP_PORT"),
                  os.getenv("APP_HOST"),
                  os.getenv("OUTPUT_WRITER_PORT"),
                  os.getenv("OUTPUT_WRITER_HOST"))
    app.start()

    # setup output
    ow_config = {
        'OUTPUT_WRITER_PORT': os.getenv("OUTPUT_WRITER_PORT"),
        'OUTPUT_WRITER_HOST': os.getenv("OUTPUT_WRITER_HOST"),
    }

    print('Creating writer')
    output_client_mock = StreamingClientMock()
    ow = OutputWriterFactory.create(ow_config, output_client_mock, None)
    print('start_incoming_messages')

    # Set to fail on Output Send:
    output_client_mock.set_fail_send(True)
    print('sending test message to reader')
    test_msg = str(time.clock())
    # send a message from INPUT reader, and expect it to flow in the pipeline,
    # and eventually be picked up by the output writer
    result = input_client_mock.fake_incoming_message_from_streaming(test_msg)
    last_msg = output_client_mock.get_last_msg()

    assert last_msg is None
    assert result is False
    ir.stop_incoming_messages()
    ow.stop_incoming_messages()


@pytest.mark.integration
def test_when_messages_sent_to_kafka_then_all_messages_are_sent_via_output():
    """
    This function tests the integration of input and output.
    It assumes connectivity to streaming client is correct, therefore the steaming client is mocked.
    We look to find that once a message is received, it is send out by the output writer.
    This is the maximal integration test we can have in an isolated environment.
    Please allow KAFKA_TIMEOUT of 30 seconds or more
    :return:
    """

    # setup input

    ir_config = {
        'client': {
            'type': 'kafka',
            'config': {
                "KAFKA_TOPIC": os.getenv("KAFKA_TOPIC_INPUT"),
                "KAFKA_ADDRESS": os.getenv("KAFKA_ADDRESS"),
                "KAFKA_CONSUMER_GROUP": os.getenv("KAFKA_CONSUMER_GROUP"),
                "TIMEOUT": os.getenv("KAFKA_TIMEOUT")
            }
        },
        'APP_PORT': os.getenv("APP_PORT"),
        'APP_HOST': os.getenv("APP_HOST"),
    }
    print('Creating reader')
    ir = InputReaderFactory.create(ir_config)

    # setup app
    app = TestApp(os.getenv("APP_PORT"),
                  os.getenv("APP_HOST"),
                  os.getenv("OUTPUT_WRITER_PORT"),
                  os.getenv("OUTPUT_WRITER_HOST"))
    app.start()

    # setup output
    ow_config = {
        'client': {
            'type': 'kafka',
            'config': {
                "KAFKA_TOPIC": os.getenv("KAFKA_TOPIC_OUTPUT"),
                "KAFKA_ADDRESS": os.getenv("KAFKA_ADDRESS"),
                "TIMEOUT": os.getenv("KAFKA_TIMEOUT")
            }
        },
        'OUTPUT_WRITER_PORT': os.getenv("OUTPUT_WRITER_PORT"),
        'OUTPUT_WRITER_HOST': os.getenv("OUTPUT_WRITER_HOST"),
    }

    print('Creating writer')
    ow = OutputWriterFactory.create(ow_config, None, None)
    print('start_incoming_messages')
    ow.start_incoming_messages()

    print('start_receiving_messages')
    # ir.start_receiving_messages()
    t_ir = threading.Thread(name='testir', target=ir.start_receiving_messages)
    t_ir.setDaemon(True)
    t_ir.start()

    print('sending test message to reader')

    test_msg = str(time.clock())
    print("sending {} to input topic".format(test_msg))
    # send a message from INPUT reader, and expect it to flow in the pipeline,
    # and eventually be picked up by the output writer
    send_message_to_kafka(test_msg)
    last_msg = read_message_from_kafka()
    print("received {} from output topic".format(last_msg))
    assert last_msg == test_msg

    ir.stop_incoming_messages()
    ow.stop_incoming_messages()


def send_message_to_kafka(msg):
    config = {
        "KAFKA_TOPIC": os.getenv("KAFKA_TOPIC_INPUT"),
        "KAFKA_ADDRESS": os.getenv("KAFKA_ADDRESS"),
        "TIMEOUT": os.getenv("KAFKA_TIMEOUT")
    }
    kafka = KafkaStreamingClient(config)
    val = kafka.send(msg)
    return val


def read_message_from_kafka():
    config = {
        "KAFKA_TOPIC": os.getenv("KAFKA_TOPIC_OUTPUT"),
        "KAFKA_ADDRESS": os.getenv("KAFKA_ADDRESS"),
        "KAFKA_CONSUMER_GROUP": os.getenv("KAFKA_CONSUMER_GROUP"),
        "TIMEOUT": os.getenv("KAFKA_TIMEOUT")
    }
    kafka = KafkaStreamingClient(config)
    kafka.start_receiving(on_msg)

    start = datetime.datetime.now()
    timeout = int(os.getenv("KAFKA_TIMEOUT"))
    stop = False
    msg = None
    while not stop:
        # Stop loop after timeout if exists
        elapsed = datetime.datetime.now() - start
        if elapsed.seconds >= timeout:
            stop = True

        # Poll messages from topic
        if my_msg is not None:
            stop = True
            msg = my_msg

    return msg


my_msg = None


def on_msg(msg):
    global my_msg
    my_msg = msg.decode('utf-8')



# -*- coding: utf-8 -*-
""" Factory for OutputWriter """

from agogosml.common.flask_http_listener_client import FlaskHttpListenerClient
from agogosml.common.eventhub_streaming_client import EventHubStreamingClient
from agogosml.common.kafka_streaming_client import KafkaStreamingClient
from .output_writer import OutputWriter


class OutputWriterFactory:
    """Factory for OutputWriter"""

    @staticmethod
    def create(config: dict, streaming_client=None, listener_client=None):
        """Creates a new instance of OutputWriter.

        :param config: A configuration for OutputWriter.
        :param streaming_client: Optional, an existing streaming client implementation to use.
        :param listener_client: Optional, an existing listener client implementation to use.
        :return OutputWriter: An instance of an OutputWriter with a
        streaming_client and listener.
        """

        client = None

        if OutputWriterFactory.is_empty(config):
            raise Exception('''
            No config was set for the OutputWriterFactory
            ''')

        if streaming_client is None:
            if config.get("client") is None:
                raise Exception('''
                client cannot be empty
                ''')

            client_config = config.get("client")["config"]
            if config.get("client")["type"] == "kafka":
                client = KafkaStreamingClient(client_config)

            if config.get("client")["type"] == "eventhub":
                client = EventHubStreamingClient(client_config)

            if client is None:
                raise Exception('''
                Unknown client type
                ''')
        else:
            client = streaming_client

        listener = None
        if listener_client is None:
            port = config.get("OUTPUT_WRITER_PORT")
            host = config.get("OUTPUT_WRITER_HOST")

            listener = FlaskHttpListenerClient(port, host)
        else:
            listener = listener_client

        return OutputWriter(client, listener)

    @staticmethod
    def is_empty(dictionary: dict) -> bool:
        """
        Checks if a dictionary is empty.
        Empty dictionaries resolve to false when
        converted to booleans in Python.

        :param dictionary: A dictionary to test.
        :return: True if empty, false otherwise.
        """
        return not bool(dictionary)



# -*- coding: utf-8 -*-

"""Top-level package for Agogosml."""

__author__ = """Rami Sayar"""
__email__ = 'rami.sayar@microsoft.com'
__version__ = '0.1.2'

if __name__ == '__main__':
    print(__version__)



"""Abstract listener client broker class"""

from abc import ABC, abstractmethod


class ListenerClient(ABC):

    @abstractmethod
    def __init__(self, port, host):
        pass

    @abstractmethod
    def start(self, on_message_received):
        pass

    @abstractmethod
    def stop(self):
        pass



# -*- coding: utf-8 -*-
""" OutputWriter class"""

from agogosml.common.abstract_streaming_client import AbstractStreamingClient
from agogosml.common.listener_client import ListenerClient


class OutputWriter:
    """Accepts incoming messages and routes them to a configured output."""

    def __init__(self, streaming_client: AbstractStreamingClient,
                 listener: ListenerClient):
        """
        :param streaming_client: A client that can stream data in.
        :param listener: A client that accepts incoming messages.
        """
        self.messaging_client = streaming_client
        self.listener = listener

    def on_message_received(self, message):
        """Send messages onwards to a streaming client.

        :param message: A message as a string to send onwards.
        """
        return self.messaging_client.send(message)

    def start_incoming_messages(self, callback=None):
        """Start accepting messages."""
        if callback:
            self.listener.start(callback)
        else:
            self.listener.start(self.on_message_received)

    def stop_incoming_messages(self):
        """Stop accepting messages."""
        self.listener.stop()



import pytest
import os
from agogosml.writer.output_writer import OutputWriter
from agogosml.writer.output_writer_factory import OutputWriterFactory
from agogosml.common.flask_http_listener_client import FlaskHttpListenerClient
from agogosml.common.kafka_streaming_client import KafkaStreamingClient
from agogosml.common.eventhub_streaming_client import EventHubStreamingClient
from .client_mocks import StreamingClientMock, HttpClientMock


@pytest.fixture
def mock_streaming_client():
    return StreamingClientMock()


@pytest.fixture
def mock_listener_client():
    return HttpClientMock(0)


def test_when_ctor_instance_created(mock_streaming_client, mock_listener_client):
    ow = OutputWriter(mock_streaming_client, mock_listener_client)
    assert ow is not None


def test_on_message_received_sent_called(mock_streaming_client,
                                         mock_listener_client):
    ow = OutputWriter(mock_streaming_client, mock_listener_client)
    ow.on_message_received('test')
    assert mock_streaming_client.get_sent()


def test_on_listener_event_sent_called(mock_streaming_client,
                                       mock_listener_client):
    ow = OutputWriter(mock_streaming_client, mock_listener_client)
    ow.start_incoming_messages()
    mock_listener_client.mock_new_incoming_message()
    assert mock_streaming_client.get_sent()
    assert mock_listener_client.get_started()


def test_when_failed_to_send_then_report_error(mock_streaming_client, mock_listener_client):
    # arrange
    ow = OutputWriter(mock_streaming_client, mock_listener_client)
    ow.start_incoming_messages()
    mock_streaming_client.set_fail_send(True)
    # act
    result = mock_listener_client.mock_new_incoming_message()
    # assert
    assert result is False


def test_on_stop_event_stop_called(mock_streaming_client, mock_listener_client):
    ow = OutputWriter(mock_streaming_client, mock_listener_client)
    ow.start_incoming_messages()
    ow.stop_incoming_messages()
    assert mock_listener_client.get_stopped()


def test_when_unknown_client_throw():
    config = {'client': {'type': 'aaa'}}
    with pytest.raises(Exception):
        OutputWriterFactory.create(config)


@pytest.mark.integration
def test_output_writer_factory_event_hub():

    config = {
        'client': {
            'type': 'eventhub',
            'config': {
                'EVENT_HUB_NAMESPACE': os.getenv("EVENT_HUB_NAMESPACE"),
                'EVENT_HUB_NAME': os.getenv("EVENT_HUB_NAME"),
                'EVENT_HUB_SAS_POLICY': os.getenv("EVENT_HUB_SAS_POLICY"),
                'EVENT_HUB_SAS_KEY': os.getenv("EVENT_HUB_SAS_KEY"),
                'OUTPUT_WRITER_PORT': os.getenv("OUTPUT_WRITER_PORT"),
            }
        }
    }

    ow = OutputWriterFactory.create(config)
    assert ow is not None
    assert isinstance(ow.messaging_client, EventHubStreamingClient)
    ow.messaging_client.stop()


@pytest.mark.integration
def test_output_writer_factory_kafka():

    config = {
        'client': {
            'type': 'kafka',
            'config': {
                'KAFKA_TOPIC1': os.getenv("KAFKA_TOPIC"),
                'KAFKA_ADDRESS': os.getenv("KAFKA_ADDRESS"),
                'OUTPUT_WRITER_PORT': os.getenv("OUTPUT_WRITER_PORT"),
            }
        }
    }

    ow = OutputWriterFactory.create(config)
    assert ow is not None
    assert isinstance(ow.messaging_client, KafkaStreamingClient)
    ow.messaging_client.stop()


@pytest.mark.integration
def test_output_writer_flask():

    config = {
        'client': {
            'type': 'kafka',
            'config': {
                'KAFKA_TOPIC': os.getenv("KAFKA_TOPIC"),
                'KAFKA_ADDRESS': os.getenv("KAFKA_ADDRESS"),
                'OUTPUT_WRITER_PORT': os.getenv("OUTPUT_WRITER_PORT"),
                'OUTPUT_WRITER_HOST': os.getenv("OUTPUT_WRITER_HOST"),
            }
        }
    }

    ow = OutputWriterFactory.create(config)
    assert ow is not None
    assert isinstance(ow.listener, FlaskHttpListenerClient)

    received_messages = []

    def receive_callback(message):
        received_messages.append(message)

    ow.listener.start(receive_callback)



"""HttpMessageSender."""

import requests
from .message_sender import MessageSender
from ..utils.logger import Logger

logger = Logger()


class HttpMessageSender(MessageSender):  # pylint: disable=too-few-public-methods
    """HttpMessageSender."""

    def __init__(self, host_endpoint, port_endpoint):
        """

        :param host_endpoint: Host endpoint.
        :param port_endpoint: Port number.
        """
        logger.info("host_endpoint: {}".format(host_endpoint))
        logger.info("port_endpoint: {}".format(port_endpoint))

        if host_endpoint is None:
            raise ValueError('Host endpoint cannot be None.')

        if host_endpoint == "":
            raise ValueError('Host endpoint cannot be empty.')

        if int(port_endpoint) <= 0:
            raise ValueError('Port cannot be 0 or less.')

        self.host_endpoint = host_endpoint
        self.port_endpoint = port_endpoint

    def send(self, message):
        """
        Sends messages to specified address via HTTP.

        :param message: JSON formatted message.
        """
        return_value = False
        try:
            server_address = "http://" + self.host_endpoint + ":" + self.port_endpoint
            # TODO: Add retries as some of the messages are failing to send
            request = requests.post(server_address, data=message)
            if request.status_code != 200:
                logger.error(
                    "Error with a request {} and message not sent was {}".
                    format(request.status_code, message))
                print("Error with a request {} and message not sent was {}".
                      format(request.status_code, message))
            else:
                return_value = True

        except Exception as e_e:
            logger.error('Failed to send request: ' + str(e_e))

        return return_value



""" Logger """
import os
import logging.config
import yaml


class Logger(object):
    """A logger implementation."""

    __instance = None
    logger = None

    @staticmethod
    def setup_logging(log_path='logging.yaml',
                      level=logging.INFO,
                      env_key='LOG_CFG'):
        """Setup logging configuration."""

        path = log_path
        value = os.getenv(env_key, None)
        if value:
            path = value
        if os.path.exists(path):
            with open(path, 'rt') as f:
                config = yaml.safe_load(f.read())
            logging.config.dictConfig(config)
        else:
            logging.basicConfig(
                format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                level=level)

    def __new__(cls, *args, **kwargs):
        if not cls.__instance:
            cls.__instance = object.__new__(cls, *args, **kwargs)
        return cls.__instance

    def __init__(self,
                 name=__name__,
                 path='logging.yaml',
                 env_key='LOG_CFG',
                 level=logging.INFO):

        self.level = level
        self.name = name
        self.path = path
        self.env_key = env_key
        self.logger = None

        Logger.setup_logging(
            log_path=self.path, level=self.level, env_key=self.env_key)

        self.logger = logging.getLogger(self.name)

    def debug(self, message):
        """
        Log debug message.

        :param message: Debug message string.
        """
        self.logger.debug(message)

    def info(self, message):
        """
        Log info message

        :param message: Info message string.
        """
        self.logger.info(message)

    def error(self, message):
        """
        Log error message

        :param message: Error message string.
        """
        self.logger.error(message)



"""Event Hub streaming client"""

import asyncio
from azure.eventhub import EventHubClient, EventData
from azure.eventprocessorhost import AzureStorageCheckpointLeaseManager, \
    EventHubConfig, EventProcessorHost, EPHOptions
from .eventhub_processor_events import EventProcessor
from .abstract_streaming_client import AbstractStreamingClient
from ..utils.logger import Logger

logger = Logger()


class EventHubStreamingClient(AbstractStreamingClient):
    def __init__(self, config):
        """
        Class to create an EventHubStreamingClient instance.

        :param config: Dictionary file with all the relevant parameters.
        """
        super().__init__()
        self.message_callback = None
        self.config = config
        self.storage_account_name = self.config.get("AZURE_STORAGE_ACCOUNT")
        self.storage_key = self.config.get("AZURE_STORAGE_ACCESS_KEY")
        self.lease_container_name = self.config.get("LEASE_CONTAINER_NAME")
        self.namespace = self.config.get("EVENT_HUB_NAMESPACE")
        self.eventhub = self.config.get("EVENT_HUB_NAME")
        self.consumer_group = self.config.get("EVENT_HUB_CONSUMER_GROUP")
        if self.consumer_group is None:
            self.consumer_group = '$default'

        self.user = self.config.get("EVENT_HUB_SAS_POLICY")
        self.key = self.config.get("EVENT_HUB_SAS_KEY")
        if self.config.get("TIMEOUT"):
            try:
                self.timeout = int(self.config.get("TIMEOUT"))
            except ValueError:
                self.timeout = None
        else:
            self.timeout = None

        # Create EPH Client
        if self.storage_account_name is not None and self.storage_key is not None:
            self.eph_client = EventHubConfig(
                sb_name=self.namespace,
                eh_name=self.eventhub,
                policy=self.user,
                sas_key=self.key,
                consumer_group=self.consumer_group)
            self.eh_options = EPHOptions()
            self.eh_options.release_pump_on_timeout = True
            self.eh_options.auto_reconnect_on_error = False
            self.eh_options.debug_trace = False
            self.storage_manager = AzureStorageCheckpointLeaseManager(
                self.storage_account_name, self.storage_key,
                self.lease_container_name)

        # Create Send client
        else:
            address = "amqps://" + self.namespace + \
                      ".servicebus.windows.net/" + self.eventhub
            try:
                self.send_client = EventHubClient(
                    address,
                    debug=False,
                    username=self.user,
                    password=self.key)
                self.sender = self.send_client.add_sender()
                self.send_client.run()
            except Exception as e:
                logger.error('Failed to init EH send client: ' + str(e))
                raise

    def start_receiving(self, on_message_received_callback):
        """
        Receive messages from an EventHubStreamingClient.

        :param on_message_received_callback: Callback function.
        """
        loop = asyncio.get_event_loop()
        try:
            host = EventProcessorHost(
                EventProcessor,
                self.eph_client,
                self.storage_manager,
                ep_params=[on_message_received_callback],
                eph_options=self.eh_options,
                loop=loop)

            tasks = asyncio.gather(host.open_async(),
                                   self.wait_and_close(host, self.timeout))
            loop.run_until_complete(tasks)

        finally:
            loop.stop()

    def send(self, message):
        """
        Send a message to an EventHubStreamingClient.

        :param message: A string input to upload to event hub.
        """
        try:
            self.sender.send(EventData(body=message))
            logger.info('Sent message: {}'.format(message))
            return True
        except Exception as e:
            logger.error('Failed to send message to EH: ' + str(e))
            return False

    def stop(self):
        try:
            self.send_client.stop()
        except Exception as e:
            logger.error('Failed to close send client: ' + str(e))

    @staticmethod
    async def wait_and_close(host, timeout):
        """
        Run EventProcessorHost indefinitely.

        :param host: An EventProcessorHost instance.
        :param timeout: Length of time to run the EventProcessorHost for.
        """
        if timeout is None:
            while True:
                await asyncio.sleep(1)

        else:
            await asyncio.sleep(timeout)
            await host.close_async()



"""Kafka streaming client"""

from confluent_kafka import Producer, Consumer, admin
from confluent_kafka import KafkaException, KafkaError
import datetime
from .abstract_streaming_client import AbstractStreamingClient
from ..utils.logger import Logger

logger = Logger()


class KafkaStreamingClient(AbstractStreamingClient):
    def __init__(self, config):
        """
        Class to create a KafkaStreamingClient instance.

        :param config: Dictionary file with all the relevant parameters.
        :param topic: A string kafka topic.
        """

        self.topic = config.get("KAFKA_TOPIC")
        if config.get("TIMEOUT"):
            try:
                self.timeout = int(config.get("TIMEOUT"))
            except ValueError:
                self.timeout = None
        else:
            self.timeout = None

        kafka_config = self.create_kafka_config(config)
        self.admin = admin.AdminClient(kafka_config)

        if config.get("KAFKA_CONSUMER_GROUP") is None:
            self.producer = Producer(kafka_config)
        else:
            self.app_host = config.get("APP_HOST")
            self.app_port = config.get("APP_PORT")
            self.consumer = Consumer(kafka_config)

    @staticmethod
    def create_kafka_config(user_config):

        config = {
            "bootstrap.servers": user_config.get("KAFKA_ADDRESS"),
            "enable.auto.commit": False,
            "auto.offset.reset": "earliest"
        }

        consumer_group = user_config.get("KAFKA_CONSUMER_GROUP")
        if consumer_group is not None:
            config["group.id"] = consumer_group

        return config

    def delivery_report(self, err, msg):
        """ Called once for each message produced to indicate delivery result.
        Triggered by poll() or flush().

        :param err: An error message.
        :param msg: A string input to be uploaded to kafka.
        """

        if err is not None:
            logger.error('Message delivery failed: {}'.format(err))
        else:
            logger.info('Message delivered to {} [{}]'.format(
                msg.topic(), msg.partition()))

    def send(self, message: str, *args, **kwargs):
        """
        Upload a message to a kafka topic.

        :param message: A string input to upload to kafka.
        """
        if not isinstance(message, str):
            raise TypeError('str type expected for message')
        try:
            mutated_message = message.encode('utf-8')
            self.producer.poll(0)
            self.producer.produce(
                self.topic, mutated_message, callback=self.delivery_report)
            self.producer.flush()
            return True
        except Exception as e:
            logger.error('Error sending message to kafka:' + str(e))
            return False

    def stop(self, *args, **kwargs):
        pass

    def check_timeout(self, start):
        """
        Checks how much time has elapsed since the kafka client started running.

        :param start: Start time.
        """
        if self.timeout is not None:
            elapsed = datetime.datetime.now() - start
            if elapsed.seconds >= self.timeout:
                raise KeyboardInterrupt

    def handle_kafka_error(self, msg):
        """
        Handle an error in kafka.

        :param msg: Error message from kafka.
        """
        if msg.error().code() == KafkaError._PARTITION_EOF:
            # End of partition event
            logger.error('%% %s [%d] reached end at offset %d\n' %
                         (msg.topic(), msg.partition(), msg.offset()))
        else:
            # Error
            raise KafkaException(msg.error())

    def start_receiving(self, on_message_received_callback):
        """
        Receive messages from a kafka topic.

        :param on_message_received_callback: Callback function.
        """
        '''
        TODO:
        We are going to need documentation for Kafka
        to ensure proper syntax is clear
        '''
        try:
            self.subscribe_to_topic()
            start = datetime.datetime.now()

            while True:
                # Stop loop after timeout if exists
                self.check_timeout(start)

                # Poll messages from topic
                msg = self.read_single_message()
                if msg is not None:
                    on_message_received_callback(msg)

        except KeyboardInterrupt:
            logger.info('Aborting listener...')

        finally:
            # Close down consumer to commit final offsets.
            self.consumer.close()

    def subscribe_to_topic(self):
        self.consumer.subscribe([self.topic])

    def read_single_message(self):
        # Poll messages from topic
        msg = self.consumer.poll(0.000001)
        if msg is None:
            return None
        if msg.error():
            # Error or event
            self.handle_kafka_error(msg)
            return None
        else:
            # Proper message
            # logger.info('kafka read message: {}, from topic: {}'.format(msg.value(), msg.topic()))
            self.consumer.commit(msg)
            return msg.value()



"""EventProcessor host class for Event Hub"""

from azure.eventprocessorhost import AbstractEventProcessor
from ..utils.logger import Logger

logger = Logger()


class EventProcessor(AbstractEventProcessor):
    """Example Implementation of AbstractEventProcessor."""

    def __init__(self, params):
        """
        Init Event processor.

        :param params: List of params.
        """
        super().__init__()
        self.on_message_received_callback = params[0]
        self._msg_counter = 0

    async def open_async(self, context):
        """
        Called by processor host to initialize the event processor.
        """
        logger.info("Connection established {}".format(context.partition_id))

    async def close_async(self, context, reason):
        """
        Called by processor host to indicate that the event processor
        is being stopped.

        :param context: Information about the partition.
        :type context: ~azure.eventprocessorhost.PartitionContext
        :param reason: Reason for closing the async loop.
        :type reason: string
        """
        logger.info(
            "Connection closed (reason {}, id {}, offset {}, sq_number {})".
            format(reason, context.partition_id, context.offset,
                   context.sequence_number))

    async def process_events_async(self, context, messages):
        """
        Called by the processor host when a batch of events has arrived.
        This is where the real work of the event processor is done.

        :param context: Information about the partition.
        :type context: ~azure.eventprocessorhost.PartitionContext
        :param messages: The events to be processed.
        :type messages: list[~azure.eventhub.common.EventData]
        """
        for message in messages:
            message_json = message.body_as_str(encoding='UTF-8')
            if self.on_message_received_callback is not None:
                self.on_message_received_callback(message_json)
                logger.debug("Received message: {}".format(message_json))
        logger.info("Events processed {}".format(context.sequence_number))
        await context.checkpoint_async()

    async def process_error_async(self, context, error):
        """
        Called when the underlying client experiences an error while receiving.
        EventProcessorHost will take care of recovering from the error and
        continuing to pump messages, so no external action is required.

        :param context: Information about the partition.
        :type context: ~azure.eventprocessorhost.PartitionContext
        :param error: The error that occured.
        """
        logger.error("Event Processor Error {!r}".format(error))



# -*- coding: utf-8 -*-
""" Factory for InputReader """
from agogosml.common.eventhub_streaming_client import EventHubStreamingClient
from agogosml.common.http_message_sender import HttpMessageSender
from agogosml.common.kafka_streaming_client import KafkaStreamingClient
from agogosml.utils.logger import Logger
from .input_reader import InputReader

logger = Logger()


class InputReaderFactory:
    """Factory and instance resolving for input reader"""

    @staticmethod
    def create(config: dict, streaming_client=None):
        """Create a new instance of InputReader

        :param config: A configuration for InputReader
        :param streaming_client: Optional, an existing streaming client implementation to use.
        :return InputReader: An instance of an InputReader with streaming_client and message_sender
        """
        if InputReaderFactory.is_empty(config):
            raise Exception('''
            No config were set for the InputReader manager
            ''')

        client = None
        if streaming_client is None:
            if config.get("client") is None:
                raise Exception('''
                client cannot be empty
                ''')

            client_config = config.get("client")["config"]
            if config.get("client")["type"] == "kafka":
                client = KafkaStreamingClient(client_config)

            if config.get("client")["type"] == "eventhub":
                client = EventHubStreamingClient(client_config)

            if client is None:
                raise Exception('''
                Unknown client type
                ''')
        else:
            client = streaming_client

        # host and port from the client
        app_host = config.get("APP_HOST")
        app_port = config.get("APP_PORT")

        msg_sender = HttpMessageSender(app_host, app_port)

        return InputReader(client, msg_sender)

    @staticmethod
    def is_empty(dictionary: dict) -> bool:
        """
        Checks if a dictionary is empty.
        Empty dictionaries resolve to false when
        converted to booleans in Python.

        :param dictionary: a dictionary to test
        :return: true if empty, false otherwise
        """
        return not bool(dictionary)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "confluent-kafka==0.11.6",
    "azure-eventhub==1.2.0",
    "requests==2.20.1",
    "python-dotenv==0.9.1",
    "flask==1.0.2",
    "pyyaml>=4.2b1",
    "jsonschema==2.6.0"
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="Rami Sayar",
    author_email='rami.sayar@microsoft.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7'
    ],
    description="Agogosml Library",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='agogosml',
    name='agogosml',
    packages=find_packages(include=['agogosml**']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/Microsoft/agogosml',
    version='0.1.2',
    zip_safe=False,
)
